package Env;

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  This software is being made freely available without conditions on its
#  use or distribution.  The software was developed in part with funding
#  from the U.S. Department of Energy under contract W-31-109-Eng-38. 
#  No liability is assumed for any flaws in the software or for failure
#  of the software to work as expected.  If you find this software useful
#  in your research, we would appreciate your adding the following
#  acknowledgment:
#  
#       ``We gratefully acknowledge use of the WIT2 software developed by
#         researchers in the Mathematics and Computer Science Division of
#         Argonne National Laboratory.''
#
#---------------------------------------------------------------------------

require Exporter;
@ISA = (Exporter);

@EXPORT = qw(
	     default_wit_env
	     read_taxonomy
	     restore_wit_env
	     restore_wit_queries
	     save_wit_env
	     save_wit_queries
	    );

sub default_wit_env {

#  IN: Nothing
# OUT: %wit_env - hash with global variables

    my %wit_env;

    $wit_env{"USER"} = "";

    @{ $wit_env{'ORGS'} } = split ",", "AA,BB,DR,CY,PG,CT,PN,BS,CA,EF,MP,MT,CJ,NG,TP,HP,RC,PA,EC,YP,HI,AG,TH,MJ,PH,SC,CE"; 

    $wit_env{'MAXL'}{'PROT_TABLE'} = 30;
    $wit_env{'SHOW'}{'USER_ASSIGN'} = 1;

    $wit_env{'MPW'}{'BASE_URL'} = "http://wit/MPW";
    $wit_env{'EMP'}{'BASE_URL'} = "http://wit/EMP";

    return \%wit_env;
}

sub read_taxonomy {

# IN: Taxonomy file path
# OUT: Taxonomy string

    my $taxfile = shift;
    my @taxons;
    my $taxstr;

    if (open TAX, "<$taxfile")
    {
	$taxstr = <TAX>;
	
	if ( $taxstr ) 
	{
	    $taxstr =~ s/\s*\.\s*//g;
	    @taxons = split /\s*;\s*/, $taxstr;
	    pop @taxons;
	    $taxstr = (join "; ", @taxons) . ".";
	    $taxstr =~ s/Eubacteria/Bacteria/ig;
	} 
	else 
	{
	    $taxstr = "Undefined";
	}

	close TAX;
	
	return $taxstr;
    }
}

sub restore_wit_env {

# IN: $envdir  - directory of environment files
#     $user    - user name (can be undefined)
#
# OUT: \%wit_env - ref to hash 

    ( $envdir, $user ) = @_;

    if ( not $user ) { $user = ""; }

    my %wit_env;
    my $saved_ors = $/;
    my $remote_host;
    my $dumpstr;

    use Data::Dumper;

    if ( defined $ENV{"REMOTE_ADDR"} ) {
	$remote_host = $ENV{"REMOTE_ADDR"};
    } else {
	$remote_host = "localhost";
    }

    # Get and read user environment file of the name
    # "user@ip-number", or just "@ip-number" (if user name 
    # is empty).

    if ( open WITENV, "$envdir/$user\@$remote_host" ) 
    {	
	undef $/;

	%wit_env = eval ( $dumpstr = <WITENV> );

	$/ = $saved_ors;

	close WITENV;
    } 
    else 
    {
	%wit_env = %{ &default_wit_env; };
    }

    return \%wit_env;
}

sub restore_wit_queries {

# IN: $dir      - directory of query files
#     $user     - user name (can be undefined)
#
# OUT: Reference to list of queries 

    my ( $dir, $user ) = @_;

    my $remote_host;
    my $file;
    my $key;
    my $saved_ors = $/;
    my @queries; 

    if ( defined $ENV{"REMOTE_ADDR"} ) {
	$remote_host = $ENV{"REMOTE_ADDR"};
    } else {
	$remote_host = "localhost";
    }

    # Read from a file named "user@ip-number_query, or just 
    # "@ip-number_query" (if no user name given)
	
    $file = "$dir/$user\@$remote_host" . "_query";

    if (open QUERIES, "< $file") 
    {
	undef $/;

	@queries = eval ( <QUERIES> );

	$/ = $saved_ors;

	close QUERIES;

	return [ @queries ];
    } 
    else
    {
	return "";
    }
}

sub save_wit_env {

# IN: $envdir  - directory of environment files
#     $user    - user name (can be undefined)
#     %wit_env - ref to hash with environment variables
#
# OUT: Nothing, but writes a file.

    ( $envdir, $user, $wit_env ) = @_; 

    my $remote_host;
    my $envfile;
    my $key;

    use Data::Dumper;

    if ( defined $ENV{"REMOTE_ADDR"} ) {
	$remote_host = $ENV{"REMOTE_ADDR"};
    } else {
	$remote_host = "localhost";
    }

    # Write to a file named "user@ip-number", or just 
    # "@ip-number" (if no user name given)
	
    $envfile = "$envdir/$user\@$remote_host";

    if (open WITENV, "> $envfile") 
    {
	print WITENV Data::Dumper->Dump ([ $wit_env ], [ "*wit_env" ] );
	close WITENV;
	
	return 1;
    } 
    else
    {
	return 0; 
    }
}

sub save_wit_queries {

# IN: $dir      - directory of query files
#     $user     - user name (can be undefined)
#     $queriesP - ref to list of nested query hashes
#     $namesP   - ref to list of query names
#
# OUT: nothing

    my ( $dir, $user, $queriesP, $namesP ) = @_;

    my $remote_host;
    my $file;
    my $key;

    use Data::Dumper;

    if ( defined $ENV{"REMOTE_ADDR"} ) {
	$remote_host = $ENV{"REMOTE_ADDR"};
    } else {
	$remote_host = "localhost";
    }

    # Write to a file named "user@ip-number_query, or just 
    # "@ip-number_query" (if no user name given)
	
    $file = "$dir/$user\@$remote_host" . "_query";

    if (open QUERIES, "> $file") 
    {
	print QUERIES Data::Dumper->Dump ( $queriesP, $namesP );
	close QUERIES;
	
	return 1;
    } 
    else
    {
	return 0;
    }
}

1;

