package Random;			# -*- perl -*-

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Contains:
#    ran_set_params  ---  set bounds and parameters of mappings
#    ran_shell       ---  repeatedly calls a function until valid RV is found
#    ran_init        ---  initialize RNG
#    ran_shuffle     ---  Bays-Durham improved version of `rand()`
#    ran_uniform     ---  uniform ``boxcar'' random variable
#    ran_gamma       ---  gamma random variable
#    ran_poisson     ---  poisson random variable
#    ran_binomial    ---  binomial random variable
#    ran_neg_binom   ---  ``negative binomial'' random variable
#    ran_gaussian    ---  gaussian RV
#    empirical_dist  ---  computes empirical cumulative dist. and density
#    empirical_stats ---  computes mean, variance, skewness, etc. of data
#---------------------------------------------------------------------------

use Exporter ();
use Messages;

@ISA    = ( Exporter );
@EXPORT = qw(
	     ran_set_params	
	     ran_verbose
	     ran_shell
	     ran_init		
	     ran_shuffle	
	     ran_uniform		
	     ran_gamma      
	     ran_poisson    
	     ran_binomial
	     ran_neg_binom
	     ran_gaussian
	     empirical_dist
	     empirical_stats
	     geometric_mean
	     median
	     moments_to_cumulants
	    );
#
#===========================================================================

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#... Common package-parameters and variables:
#
#    $ran_max_iter, 
#    $ran_min, 
#    $ran_max, 
#    $ran_mean, 
#    $ran_variance, 
#    $verbose
#---------------------------------------------------------------------------

sub ran_set_params
{
    ( $ran_max_iter, $ran_min, $ran_max, $ran_mean, $ran_variance ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Sets package parameters and variables...
#   Note that above are =global= to package, 
#   not 'my' variables to '&ran_set_params()' !!!
#---------------------------------------------------------------------------
    return;
}

sub ran_verbose
{
    ( $verbose ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Sets verbose-mode for debugging puposes...
#---------------------------------------------------------------------------
    return;
}



sub ran_shell
{
    my ( $func, @args ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Repeatedly calls '&$func' with arg-list @arg until it returns
#   a random variable between $ran_min and $ran_max; 
#   converts result to nearest integer before returning it
#---------------------------------------------------------------------------
    my ( $val, $n );
    
    $n = 0;
    do {
	((++$n) <= $ran_max_iter)  ||  wit_die "ran_shell: no hits in $ran_max_iter iterations!";
	$val = int( 0.5 + &$func( @args ) );
    } while ( ($val < $ran_min) || ($val > $ran_max) ); 

    return ( $val );
}



my( $ran_val, @ran_tbl );	#... outside scope to preserve state...
sub ran_init
{
    my ( $seed, $tbl_size ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Initialize table for Bays-Durham shuffle algorithm...
#---------------------------------------------------------------------------
    my ( $i );
    unless (defined $tbl_size) { 
	$tbl_size = 97;		#...Knuth likes this number...
    } 

    unless ( defined($seed) && $seed ) 
    {
	$seed  = time();
	$seed ^= ($seed <<  1);
	$seed ^= ($seed <<  4);
	$seed ^= ($seed <<  6);
	$seed ^= ($seed << 30);
	$seed ^= $$;
	$seed  = &reverse_bits( $seed );
	$seed ^= $$;

	print STDERR "seed used is $seed\n" if ($verbose);
    }
    srand( $seed );

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...``Warm up'' system RNG with 20 extra calls, then fill table...
#---------------------------------------------------------------------------
    for ( $i = (20 + $tbl_size); ($i >= 0); $i-- )
    {
	$ran_val = rand();
	if ( $i < $tbl_size ) { $ran_tbl[ $i ] = $ran_val; }
    }
    print STDERR  "table size = ", ($tbl_size = @ran_tbl), "\n" if ($verbose);

    $ran_val = rand();

    return( $seed );
}

sub reverse_bits
{
    my ( $m ) = @_;
    my ($i, $j, $n );
    
    $n = 0;
    for ( $i=0; $i < 32; $i++ )
    {
	$n <<= 1;
	$j   = ( $m & 1 );
	$n  |= $j; 
	$m >>= 1;
    }
    return $n;
}



sub ran_shuffle
{
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Use Bays-Durham shuffling to improve the system RNG...
#---------------------------------------------------------------------------
    my ( $index, $tbl_size );
    
    if ( ($tbl_size = @ran_tbl) == 0 ) 
    { 
	&ran_init(0); 
	$tbl_size = @ran_tbl;
    }
    
    $index   = int( $tbl_size * $ran_val );
    $ran_val = $ran_tbl[ $index ];
    $ran_tbl[ $index ] = rand;
    
    return( $ran_val );
}



sub ran_uniform
{
    my ( $umin, $umax ) = @_;
    unless (defined $umin)  { $umin = $ran_min; }
    unless (defined $umax)  { $umax = $ran_max; }
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Uniform distribution between  umin and  umax.
#      mean     = (umax + umin) / 2, 
#      variance = (umax - umin) * (umax - umin + 1) / 12
#      skewness = 0
#---------------------------------------------------------------------------
    my ( $uniform );
    
    $uniform = int( $umin  +  ( $umax - $umin + 1 ) * &ran_shuffle );

    return $uniform;
}



sub ran_gamma
{
    my ( $a ) = @_;
    unless (defined $a)  { $a = $ran_mean; }
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Gamma distribution (continuous) --- 
#      p(x) := x^(a-1) * exp(-x) / GammaFcn(a)
#      mean  = a,  variance = a,  skewness = 2 / sqrt(a)
#---------------------------------------------------------------------------
    my ( $b, $c, $u, $v, $w, $x, $y, $z, $done );

    $b = $a - 1.0;
    $c = 3.0 * ( $a - 0.25 );

    do {
	$v = 0;   
	while ( ($v == 0.0) || ($v == 1.0) ) { $v = &ran_shuffle; }
	if ( $a == 1.0 ) { return ( $x = -log($v) ); }
	$u = &ran_shuffle;

	$w = $u * ( 1.0 - $u );
	$y = sqrt( $c / $w ) * ( $u - 0.5 );
	$x = $b  + $y;
	
	$done = "";
	if ( $x > 0.0 )
	{
	    $z    = 64.0 * $w**3 * $v**2 ;

#	    print STDERR "a = $a, b = $b, c = $c\n"   if ($verbose);
#	    print STDERR "u = $u, v = $v, z = $w\n"   if ($verbose);
#	    print STDERR "x = $x, y = $y, z = $z\n\n" if ($verbose);

	    $done = ($z <= (1.0 - 2.0 * $y**2 / $x ));
	    unless ( $done ) 
	    { 
		$done = ( log($z) <= 2.0*($b*log($x/$b)-$y) );
	    }
	}
    } until ( $done );
    return $x;
}



sub ran_poisson
{
    my ( $lambda ) = @_;
    unless (defined $lambda ) { $lambda = $ran_mean; }
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Poisson distribution (discrete) --- P(n) = lambda^n * exp(-lambda) / n!
#   mean = lambda,  variance = lambda,  skewness = 1/sqrt(lambda)
#---------------------------------------------------------------------------
    my ( $g, $n, $t, $ln_lambda, $sq, $y);
    my ( $pi ) = 4.0 * atan2(1.0,1.0);

    if ( $lambda <= 12.0 ) 
    {
	$g = exp( -$lambda );
        $n = -1.0;
        $t =  1.0;
	do {
	    ++$n;
	    $t *= &ran_shuffle;
	} while ( $t > $g );
    } 
    else 
    {
	$sq = sqrt( 2.0 * $lambda );
	$ln_lambda = log( $lambda );
	$g  = $lambda * $ln_lambda - &lnGamma( $lambda + 1.0);

	do {
	    do {
		do { $y = $pi * ran_shuffle; } until ( $y != 0.0 );
		$y = cos( $y );
		$y = sqrt( 1.0 - $y**2 ) / $y;
		$n = $sq * $y + $lambda;
	    } while ( $n < 0.0 );
	    $n = int( $n );
	    $t  = 0.9 * (1.0 + $y**2) 
		* exp( $n * $ln_lambda - &lnGamma($n + 1.0) - $g);
	} while (&ran_shuffle > $t);
    }
    return $n;
}



sub lnGamma
{
    my ( $x ) = my ( $xx ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Natural logarithm of the Gamma function  ---  n! == Gamma(n+1)
#---------------------------------------------------------------------------
    my ( $j, $tmp );

    my ( $stp ) = ( 2.5066282746310005 );
    my ( $ser ) = ( 1.000000000190015  );
    my ( @cof ) = (  76.18009172947146,    -86.50532032032941677
		   , 24.01409824083091,     -1.231739572450155
		   ,  1.208650973866179e-3, -5.3695239384953e-6  );
    
    if ( $x <= 0 ) { return undef(); }

    $tmp  =  $x + 5.5;
    $tmp -= ($x + 0.5) * log( $tmp );

    for ( $j=0; ($j < 6); $j++ )
    {
	$ser += $cof[$j] / (++$x);
    }

    return ( log( $stp * $ser / $xx ) - $tmp );
}



my ( $bnldev );			# 
my ( $nold ) = -1;		#...must be outside scope to preserve state
my ( $pold ) = -1;		# 
sub ran_binomial
{
    my ( $n, $pp ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Generates binomial (discrete) random variates drawn from distribution:
#        P(k) = Binomial_coeff(n,k) * p^k * q^(n-k),  q := (1-p).
#   mean = n*p,  variance = n*p*q,  skewness = (q - p) / sqrt( n*p*q ).
#   Result 'k' is in the range (0 <= k <= n)
#
#...If (n,p) are unspecified, routine will compute them to match the
#   mean to $ran_mean, and the variance to $ran_variance; however, 
#   this is only done on the FIRST CALL, so remember to call the
#   routine with argument "reset" if $ran_mean and $ran_variance change !!!
#---------------------------------------------------------------------------
    my ( $p, $q, $plog, $qlog, $en, $oldg );
    my ( $am, $em, $g, $sq, $t, $y, $j );

    my ( $pi ) = 4.0 * atan2(1,1);

    if (defined $n) 
    { 
	if ( $n eq "reset" ) { return( undef($bnldev) ); }

	($n >= 1 )                ||  wit_die "ran_binomial: n = $n < 1";
	(($pp > 0) && ($pp < 1))  ||  wit_die "ran_binomial: p = $pp not between 0 and 1";
    } 
    else 
    {
	unless ( $ran_variance < $ran_mean ) 
	{
	    wit_die "ran_binom: variance = $ran_variance must be less than mean = $ran_mean";
	}
	$n = int( 0.5 + $ran_mean**2 / ($ran_mean - $ran_variance) );
	$n = ($n > 1) ? $n : 1 ;
	$p = $pp = $ran_mean / $n;
	$q = (1.0 - $pp);
    }

    unless (defined $bnldev )
    {
	warn "initializing ran_binomial: p = $p, q = $q, n = $n\n";
	warn "expected mean = ", ($n * $p)
           , ", variance = ", ($n * $p * $q)
           , ", skewness = ", ($q - $p) / sqrt($n * $p * $q), "\n";
    } 

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...binomial dist. is invariant under transformation:
#        p <--> (1-p),  bnldev <--> (n - bnldev)
#   We perform the latter just before returning $bnldev.
#---------------------------------------------------------------------------
    $p = ( $pp <= 0.5 ) ? $pp : (1.0 - $pp); 
    
    $am = $n * $p;
    if ($n < 25) 
    {
	$bnldev = 0.0;
	for ( $j=1; ($j <= $n); ++$j) 
	{
	    if ( &ran_shuffle < $p )  { ++$bnldev; }
	}
    } 
    elsif ($am < 1.)  
    {
        $g = exp(-$am);
        $t = 1.0;
	for ( $j=0; ($j <= $n); ++$j) 
	{
	    $t *= &ran_shuffle;
	    last if ($t < $g);
	}
	$bnldev = ($j <= $n) ? $j : $n;
    } else {
#       if ($n != $nold)  
	{
	    $en   = $n;
	    $oldg = &lnGamma($en + 1.);
	    $nold = $n;
	}
	
#	if ($p != $pold)  
	{
	    $q    = 1.0 - $p;
	    $plog = log($p);
	    $qlog = log($q);
	    $pold = $p;
	}
	
	do {
	    do {
		$sq = sqrt(2.0 * $am * $q);

		do { $y = $pi * ran_shuffle; } until ( $y != 0.0 );
		$y = cos( $y );
		$y = sqrt( 1.0 - $y**2 ) / $y;

		$em = $sq * $y + $am;
	    } while ( ($em < 0.0) || ($em >= $en + 1.0) );
	    $em = int($em);
	    $t = 1.2 * $sq * (1.0 + $y**2)
		 * exp( $oldg - &lnGamma($em + 1.0) 
		        - &lnGamma($en - $em + 1.0) 
		        + $em * $plog + ($en - $em) * $qlog
		      );
	} while ( &ran_shuffle > $t );
	$bnldev = $em;
    }
    
    if ($p != $pp) { $bnldev = $n - $bnldev; }
    return( $bnldev );
}



my ( $neg_binom );		#...must be outside scope to preserve state
sub ran_neg_binom
{
    my ( $r, $p ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Generalized ``waiting-time'' discrete distribution --- 
#     P(n) = Binomial_coeff( n+r-1; n ) * p^r * (1-p)^n
#       r := number of 'hits' needed to terminate 'waiting' process
#       p := probability of a 'hit';    q == (1 - p)
#     mean = r*q/p,  variance = r*q/p**2,  skewness = (1+q) / sqrt(rp)
#
#...If (r,p) are unspecified, routine will compute them to match the
#   mean to $ran_mean, and the variance to $ran_variance; however, 
#   this is only done on the FIRST CALL, so remember to call the
#   routine with argument "reset" if $ran_mean and $ran_variance change !!!
#---------------------------------------------------------------------------
    my ( $q );

    if (defined $r)
    {
	if ( $r eq "reset" ) { return( undef($neg_binom) ); }

	($r >= 1 )              ||  wit_die "ran_neg_binom: r = $r < 1";
	(($p > 0) && ($p < 1))  ||  wit_die "ran_neg_binom: p = $p not between 0 and 1";

	$q = (1.0 - $p);
    } 
    else 
    {
	unless ( $ran_variance > $ran_mean )
	{
	    wit_die "ran_neg_binom: variance = $ran_variance must exceed mean = $ran_mean";
	}
	
	$r = ( $ran_mean ** 2 ) / ( $ran_variance - $ran_mean );
	$r = ( $r < 1 ) ? 1 : $r;       #...this is a hack...
	$p = $r / ( $r + $ran_mean );
	$q = (1 - $p);
    }

    unless (defined $neg_binom )
    {
	warn "initializing ran_neg_binom: p = $p, q = $q, r = $r\n";
	warn "expected mean = ", ($r*($q/$p))
           , ", variance = ", ($r*($q/$p**2))
           , ", skewness = ", ((1.0+$q) / sqrt($r*$q)), "\n";
    } 

    $neg_binom = &ran_poisson( ($q/$p) * &ran_gamma($r) );
    return( $neg_binom );
}



my ( $gsave );			#...must be outside scope to preserve state
sub ran_gaussian
{
    my ( $mean, $variance ) = @_;
    unless (defined $mean)     { $mean = $ran_mean; }
    unless (defined $variance) { $variance = $ran_variance; }
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Generates pair of Gaussian random vars via Box-Muller algorithm;
#   returns first, and saves second for next call...
#---------------------------------------------------------------------------
    my ( $u, $v, $rsq, $rmap, $gauss );
    
    if (defined $gsave)		#...spare unit-gauss RV left from last call
    {
	$gauss = $gsave;  
	undef  ( $gsave );
    }
    else			#...generate new pair of RVs by Box-Muller
    {
	do {
	    $u = 2.0 * &ran_shuffle - 1.0;
	    $v = 2.0 * &ran_shuffle - 1.0;

	    $rsq = $u**2 + $v**2;
	} while ( ($rsq > 1.0) || ($rsq == 0.0) );

	$rmap  = sqrt( -2.0 * log($rsq) / $rsq );
	$gsave = $rmap * $u;
	$gauss = $rmap * $v;
    }
    $gauss = $mean + $gauss * sqrt( $variance );
    return  ( $gauss );
}



sub empirical_dist
{
    my ( $width, $num_quantiles, $samples, $already_sorted ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Computes empirical estimates of the cumulative distribution function
#   and the probability density; also produces a sort of the data.
#   Note that the data are passed in via a pointer; the empirical
#   distributions and sorted data are also passed back as pointers...
#---------------------------------------------------------------------------
    my ( $x, $tmp, $tot,  $pop_size, $i, $index, $quantile );
    my ( $i_minus, $x_minus, $y_minus );
    my ( $i_plus,  $x_plus,  $y_plus  );

    my ( $sorted, $cumulative, $density );
    
    if ( defined($already_sorted) && $already_sorted )
    {
	$sorted  = $samples;
    } else {
	print STDERR "Random::empirical_dist --- sorting samples...\n" if ($verbose);
	$sorted  = [];
	@$sorted = sort {$a <=> $b} @$samples;
    }
    
    $tot = 0;
    $pop_size   = @$sorted;	
    $cumulative = {};
    print STDERR "Random::empirical_dist --- building cumulative distribution...\n" if ($verbose);
    foreach $x ( @$sorted )
    {
	$tot += (0.5) / $pop_size;
	unless ( $tmp = $cumulative->{ $x } )
	{
	    $cumulative->{ $x }  = $tot;
	} else {
	    $cumulative->{ $x } += (0.5) / $pop_size;
	}
	$tot += (0.5) / $pop_size;
    }

    $tot = 0;
    $index    = 0;
    $quantile = 1;
    $density  = {};
    print STDERR "Random::empirical_dist --- building density...\n" if ($verbose);
    foreach $x ( @$sorted )
    {
	$tot +=  (0.5) / $pop_size;
	if ( $tot > ($quantile/$num_quantiles) )
	{
	    $x_minus = $x;
	    $i_minus = $index - $width - 1;
	    for ( 
		  $i_minus = ( $i_minus >= 0) ? $i_minus : 0
	        ; (($x_minus == $x) && ( $i_minus >= 0 ))
		; $i_minus--
		) 
	    { 
		$x_minus = $sorted->[ $i_minus ]; 
#		warn "index = $index,\ti- = $i_minus,\tx- = $x_minus\n";
	    }
#	    warn "\n";

	    $x_minus = $sorted->[     $i_minus ];
	    $y_minus = $cumulative->{ $x_minus };
	    
	    $x_plus = $x;
	    $i_plus = $index + $width;
	    for ( 
		  $i_plus = ( $i_plus < $pop_size) ? $i_plus : $pop_size-1
		; (($x_plus == $x) && ( $i_plus < $pop_size ))
		; $i_plus++
                ) 
	    { 
		$x_plus = $sorted->[ $i_plus ];
#		warn "index = $index,\ti+ = $i_plus,\tx+ = $x_plus\n";
	    }
#	    warn "\n\n";

	    $x_plus  = $sorted->[     $i_plus ];
	    $y_plus  = $cumulative->{ $x_plus };
	    
	    if ( ($x_plus - $x_minus ) != 0.0 )
	    {
		$density->{$x} = 
		    [$index, (($y_plus - $y_minus)/($x_plus - $x_minus))];
	    } else {
		$density->{$x} = undef();
	    }
	    ++$quantile;
	}
	$tot +=  (0.5) / $pop_size;

	++$index;
    }
    
    return ( $cumulative, $density, $sorted );
}



sub empirical_stats
{
    my ( $samples, $order ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Computes mean, variance and skewness of data; note that a pointer 
#   to the data-array is passed in, rather than the data itself...
#   $n is the order of the highest normalized moment desired (default = 2)
#---------------------------------------------------------------------------
    my ( $two_pi ) = 8.0 * atan2(1.0,1.0);
    my ( $x, $pop_size, );
#   my ( $sample_mean, $sample_variance, $sample_skewness );
    my ( @moments );
    
    $pop_size = @$samples;
    if (! defined($order))  { $order = 2; }
    
    print STDERR "Random::empirical_stats called with pop_size = $pop_size, order = $order\n"  if ($verbose);
    
    for ( $n = 1; $n <= $order; $n++ )  { $moments[$n-1] = 0; }
    
    foreach $x ( @$samples )  { $moments[0] += $x; }
    $moments[0] /= $pop_size;
    
#   $sample_variance  = 0;
#   $sample_skewness  = 0;
    foreach $x ( @$samples )
    { 
#	$sample_variance += ( $x - $moments[0] ) ** 2;
#	$sample_skewness += ( $x - $moments[0] ) ** 3; 
	
	$prod = $diff = ( $x - $moments[0] );
	for ( $n = 2; $n <= $order; $n++ )
	{
	    $moments[$n-1] += (	$prod *= $diff );
	}
    }
    
#   $sample_variance /= ($pop_size-1.0);
#   
#   $sample_skewness /= ($pop_size-1.0) * ($pop_size-2.0) / $pop_size;
#   $sample_skewness /= $sample_variance ** 1.5;
    
#   print STDERR "mean  = $moments[0]\n" if ($verbose);
#   if ( $moments[1] > 0 )
#   {
#      $sigma = sqrt( $moments[1] / ($pop_size-1.0) );
#      print STDERR "sigma = $sigma\n" if ($verbose);
#   }
#   else
#   {
#      print STDERR "!!! sigma vanishes in Random::empirical_stats !!!\n" if ($verbose);
#   }
    
    
    
#   print STDERR "\n" if ($verbose);
    for ( $n = 2; $n <= $order; $n++ )
    {
 	$moments[$n-1] /= $pop_size;
	
#...This is an odd normalization, but has the advantage that 
#   all the even moments of a gaussian should be unity...
#       $moments[$n-1] /= ( 2.0 ** (($n+1.0)/2.0) * $sigma ** $n 
#		            * exp(lnGamma(($n+1.0)/2.0))
#	                  ) / sqrt( $two_pi );
    }
    
    return ( @moments );
}

sub geometric_mean
{
    my ( $scores ) = @_;
    my ( @my_scores, $log_mean, $n, $num_scores, $geo_mean );
    
    @my_scores = ();
    for ($n=0; $n < @$scores; $n++)
    {
	$x = $scores->[$n];
	if ( $x <= 0 )
	{ 
	    print STDERR "Random::geometric_mean --- skipping non-positive score $x\n" if ($verbose);
	} else {
	    push( @my_scores, $x );
	}
    }
    
    $log_mean = 0;
    $num_scores = (scalar @my_scores);
    foreach $x ( @my_scores )  { $log_mean += log($x); }
    $log_mean /= $num_scores;
    $geo_mean = exp( $log_mean );

    return $geo_mean;
}

sub median
{
    my ( $samples ) = @_;
    my ( $num, $width, $dist, $dens, $sorted, $num_keys, $x, $y, $p, $q, $median );
    
    $num   = (scalar @$samples);
    $width = int(sqrt( 1 + $num ));
    print STDERR "Random::median --- num=$num, width=$width\n" if ($verbose);
    ( $dist, $dens, $sorted )  = &empirical_dist( $width, 1, $samples );
    
    @sorted_keys = sort {$a <=> $b} keys %$dist;
    $num_keys = (scalar @sorted_keys);
    wit_die  "Random::median --- FATAL ERROR: num=$num, num_keys=$num_keys\n" if ($num_keys == 0);
    warn "Random::median --- WARNING: num=$num, num_keys=$num_keys\n"     if ($num_keys <= 3);
    
    if ($num_keys == 1) { return $sorted_keys[0]; }
    
    for ($j=0; $j < $num_keys; $j++ )
    {
	$y = $sorted_keys[$j];
	$q = $dist->{$y};
	last if ( $q >= 0.5 );
    }
    
    if ( $q == 0.5 ) { return $y; }
    
    for ($i=($j-1); $i >= 0; $i-- )
    {
	$x = $sorted_keys[$i];
	$p = $dist->{$x};
	last if ( $p < $q );
    }
    if ( ($p == $q) || ($i == $j) ) { wit_die "Random::median --- FATAL ERROR: i=$i, j=$j, x=$x, y=$y, p=$p, q=$q\n"; }
    
    $median = $x + ((0.5 - $p)/($q - $p)) * ($y - $x);
    
    return $median;
}



sub moments_to_cumulants
{
    my ( @moments ) = @_;
    my ( $mean );
    my ( $cumulants );
    
    use DA;
    
    $mean = shift( @moments );
    unshift( @moments, (1.0, 0.0) );
    
    $cumulants = &DA_poke( &DA_to_T(&DA_log( &T_to_DA(\@moments))), 1, $mean );
    shift @$cumulants;
    
    return @$cumulants;
}

1;
#######  End package Random  ###########################################
